package cn.flyrise.iot.protocol;

import cn.flyrise.iot.protocol.agree.coap.PaiCoapDTLSDeviceMessageCodec;
import cn.flyrise.iot.protocol.agree.coap.PaiCoapDeviceMessageCodec;
import cn.flyrise.iot.protocol.agree.http.PaiHttpAuthenticator;
import cn.flyrise.iot.protocol.agree.http.PaiHttpDeviceMessageCodec;
import cn.flyrise.iot.protocol.agree.mqtt.PaiMqttAuthenticator;
import cn.flyrise.iot.protocol.agree.mqtt.PaiMqttDeviceMessageCodec;
import cn.flyrise.iot.protocol.agree.websocket.PaiWebsocketDeviceMessageCodec;
import org.jetlinks.core.ProtocolSupport;
import org.jetlinks.core.defaults.CompositeProtocolSupport;
import org.jetlinks.core.message.codec.DefaultTransport;
import org.jetlinks.core.metadata.DefaultConfigMetadata;
import org.jetlinks.core.metadata.DeviceConfigScope;
import org.jetlinks.core.metadata.types.EnumType;
import org.jetlinks.core.metadata.types.PasswordType;
import org.jetlinks.core.metadata.types.StringType;
import org.jetlinks.core.spi.ProtocolSupportProvider;
import org.jetlinks.core.spi.ServiceContext;
import org.jetlinks.supports.official.JetLinksDeviceMetadataCodec;
import reactor.core.publisher.Mono;

/**
 * 物联协议
 * Flyrise-MQTT
 * @author zhangqiang
 * @since 0.1
 */
public class PaiProtocolSupportProvider implements ProtocolSupportProvider {


    @Override
    public void dispose() {
        //协议卸载时执行
    }

    /**
     * MQTT 配置信息
     */
    private static final DefaultConfigMetadata mqttConfig = new DefaultConfigMetadata(
        "MQTT认证配置"
        , "MQTT客户端clentId为设备id")
        .add("username", "username", "MQTT用户名", StringType.GLOBAL)
        .add("password", "password", "MQTT密码", PasswordType.GLOBAL)
        .add("clientid", "clientid", "MQTT客户端ID", StringType.GLOBAL)
        .add("host", "host", "设备host地址", StringType.GLOBAL);


    /**
     * COAP 配置信息
     */
    private static final DefaultConfigMetadata coapConfig = new DefaultConfigMetadata(
            "CoAP认证配置",
            "使用CoAP进行数据上报时,需要对数据进行加密:" +
                    "encrypt(payload,secureKey);")
            .add("encAlg", "加密算法", "加密算法", new EnumType()
                    .addElement(EnumType.Element.of("AES", "AES加密(ECB,PKCS#5)", "加密模式:ECB,填充方式:PKCS#5")), DeviceConfigScope.product)
            .add("secureKey", "密钥", "16位密钥KEY", new PasswordType());

    /**
     * COAPDTLS 配置信息
     */
    private static final DefaultConfigMetadata coapDTLSConfig = new DefaultConfigMetadata(
            "CoAP DTLS配置",
            "使用CoAP DTLS 进行数据上报需要先进行签名认证获取token.\n" +
                    "之后上报数据需要在Option中携带token信息. \n" +
                    "自定义Option: 2110,sign ; 2111,token ")
            .add("secureKey", "密钥", "认证签名密钥", new PasswordType());


    /**
     * MQTT 配置信息
     */
    private static final DefaultConfigMetadata httpConfig = new DefaultConfigMetadata(
            "HTTP认证配置"
            , "设备请求Authorization为设备id")
            .add("authorization", "authorization", "授权信息", StringType.GLOBAL);


    @Override
    public Mono<? extends ProtocolSupport> create(ServiceContext context) {
        CompositeProtocolSupport support = new CompositeProtocolSupport();

        // 编解码协议基础信息
        support.setId("pai-official-protocol");
        support.setName("pai-official-protocol");
        support.setDescription("pai-official-protocol Version 1.0");
        // 固定为JetLinksDeviceMetadataCodec
        support.setMetadataCodec(new JetLinksDeviceMetadataCodec());

        // 如下配置支持的协议（agree目录下开发的协议） 支持协议: MQTT 、COAP 、COAPDTLS 、HTTP

        // MQTT ===========================================================================================================
        {
            // MQTT认证策略
            support.addAuthenticator(DefaultTransport.MQTT, new PaiMqttAuthenticator());
            support.addAuthenticator(DefaultTransport.MQTT_TLS, new PaiMqttAuthenticator());
            // MQTT需要的配置信息
            support.addConfigMetadata(DefaultTransport.MQTT, mqttConfig);
            support.addConfigMetadata(DefaultTransport.MQTT_TLS, mqttConfig);
            // MQTT消息解码器
            support.addMessageCodecSupport(new PaiMqttDeviceMessageCodec(DefaultTransport.MQTT));
            support.addMessageCodecSupport(new PaiMqttDeviceMessageCodec(DefaultTransport.MQTT_TLS));
        }

        // COAP ===========================================================================================================
        {
            // COAP认证策略 (COAP使用数据加解密认证,CoAP DTLS使用签名认证)
            // COAP需要的配置信息
            support.addConfigMetadata(DefaultTransport.CoAP, coapConfig);
            support.addConfigMetadata(DefaultTransport.CoAP_DTLS, coapDTLSConfig);
            // COAP消息解码器
            support.addMessageCodecSupport(new PaiCoapDeviceMessageCodec());
            support.addMessageCodecSupport(new PaiCoapDTLSDeviceMessageCodec());
        }

        // HTTP ===========================================================================================================
        {
            // HTTP认证策略
            support.addAuthenticator(DefaultTransport.HTTP, new PaiHttpAuthenticator());
            support.addAuthenticator(DefaultTransport.HTTPS, new PaiHttpAuthenticator());
            // HTTP需要的配置信息
            support.addConfigMetadata(DefaultTransport.HTTP, httpConfig);
            // HTTP消息解码器
            support.addMessageCodecSupport(new PaiHttpDeviceMessageCodec(DefaultTransport.HTTP));
            support.addMessageCodecSupport(new PaiHttpDeviceMessageCodec(DefaultTransport.HTTPS));
        }

        // TCP ===========================================================================================================
        {
            // TCP认证策略
            // TCP需要的配置信息
            // TCP消息解码器
        }

        // UDP ===========================================================================================================
        {
            // UDP认证策略
            // UDP需要的配置信息
            // UDP消息解码器
        }

        // WebSocket ======================================================================================================
        {
            // WebSocket认证策略
            // WebSocket需要的配置信息
            // WebSocket消息解码器
            support.addMessageCodecSupport(new PaiWebsocketDeviceMessageCodec());
        }

        return Mono.just(support);
    }
}
