package cn.flyrise.iot.protocol.agree.http;

import cn.flyrise.iot.protocol.topic.TopicMessageCodec;
import org.jetlinks.core.Value;
import org.jetlinks.core.defaults.Authenticator;
import org.jetlinks.core.device.*;
import org.jetlinks.core.message.codec.http.Header;
import org.jetlinks.core.message.codec.http.HttpExchangeMessage;
import org.jetlinks.pro.network.http.HttpAuthenticationRequest;
import reactor.core.publisher.Mono;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * HTTP 认证授权
 * @author zhangqiang
 * @since 0.1
 */
public class PaiHttpAuthenticator implements Authenticator {

    /**
     *  在网络连接建立的时候,可能无法获取设备的标识(如:http,websocket等),则会调用此方法来进行认证.
     *  注意: 认证通过后,需要设置设备ID.{@link AuthenticationResponse#success(String)}
     * @param request  认证请求
     * @param registry 设备注册中心
     * @return 认证结果
     */
    @Override
    public Mono<AuthenticationResponse> authenticate(@Nonnull AuthenticationRequest request, @Nonnull DeviceRegistry registry) {
        if (request instanceof HttpAuthenticationRequest) {
             HttpAuthenticationRequest http = ((HttpAuthenticationRequest) request);
            HttpExchangeMessage message = http.getHttpExchangeMessage();
            String[] topics = TopicMessageCodec.removeProductPath(message.getPath());

            return registry
                    .getDevice(topics[1])
                    .flatMap(device -> authenticate(request, device));
        }
        return Mono.just(AuthenticationResponse.error(400, "不支持的授权类型:" + request));

    }

    /**
     * 对指定对设备进行认证
     *
     * @param request 认证请求
     * @param deviceOperation  设备
     * @return 认证结果
     */
    @Override
    public Mono<AuthenticationResponse> authenticate(@Nonnull AuthenticationRequest request, @Nonnull DeviceOperator deviceOperation) {
        if (request instanceof HttpAuthenticationRequest) {
            HttpAuthenticationRequest http = ((HttpAuthenticationRequest) request);
            HttpExchangeMessage message = http.getHttpExchangeMessage();

            String[] topics = TopicMessageCodec.removeProductPath(message.getPath());
            if(!deviceOperation.getDeviceId().equals(topics[1]))return Mono.just(AuthenticationResponse.error(400, "请求设备deviceId不存在"));

            return deviceOperation.getConfigs("authorization")
                    .flatMap(values -> {
                        Header header_authorization = message.getHeader("Authorization").orElse(null);
                        if(header_authorization != null){
                            String[] authorizationValues = header_authorization.getValue();
                            String auth = values.getValue("authorization").map(Value::asString).orElse(null);
                            if(authorizationValues!=null
                                    && authorizationValues.length>0
                                    && authorizationValues[0].equals(auth))return Mono.just(AuthenticationResponse.success());
                        }
                        return Mono.just(AuthenticationResponse.error(400, "Authorization错误"));
                    });
        }
        return Mono.just(AuthenticationResponse.error(400, "不支持的授权类型:" + request));
    }
}
