package cn.flyrise.iot.protocol.topic;

import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.SneakyThrows;
import org.hswebframework.web.bean.FastBeanCopier;
import org.jetlinks.core.message.*;
import org.jetlinks.core.message.event.EventMessage;
import org.jetlinks.core.message.firmware.*;
import org.jetlinks.core.message.function.FunctionInvokeMessage;
import org.jetlinks.core.message.function.FunctionInvokeMessageReply;
import org.jetlinks.core.message.property.*;
import org.jetlinks.core.message.state.DeviceStateCheckMessage;
import org.jetlinks.core.message.state.DeviceStateCheckMessageReply;
import org.jetlinks.core.utils.TopicUtils;
import org.reactivestreams.Publisher;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.Arrays;
import java.util.Collections;
import java.util.Optional;

/**
 * 标准编解码器
 *
 * 包含：码库、编码、解码
 * 码库：产品标识+设备标识+码库特定义
 * 如：属性上报： 产品标识+设备标识+ /properties/report
 *
 * 注释：
 * A.默认为标准物模型数据格式，这直接 payload.toJavaObject(type) 转换
 * B.非标准物模型数据格式，则重写相应的doEncode、doDecode方法，进行标准物模型数据格式转换
 *
 *
 * <pre>
 *     下行Topic:
 *          读取设备属性: /{productId}/{deviceId}/properties/read
 *          修改设备属性: /{productId}/{deviceId}/properties/write
 *          调用设备功能: /{productId}/{deviceId}/function/invoke
 *
 *          //网关设备
 *          读取子设备属性: /{productId}/{deviceId}/child/{childDeviceId}/properties/read
 *          修改子设备属性: /{productId}/{deviceId}/child/{childDeviceId}/properties/write
 *          调用子设备功能: /{productId}/{deviceId}/child/{childDeviceId}/function/invoke
 *
 *      上行Topic:
 *          读取属性回复: /{productId}/{deviceId}/properties/read/reply
 *          修改属性回复: /{productId}/{deviceId}/properties/write/reply
 *          调用设备功能: /{productId}/{deviceId}/function/invoke/reply
 *          上报设备事件: /{productId}/{deviceId}/event/{eventId}
 *          上报设备属性: /{productId}/{deviceId}/properties/report
 *          上报设备派生物模型: /{productId}/{deviceId}/metadata/derived
 *
 *          //网关设备
 *          子设备上线消息: /{productId}/{deviceId}/child/{childDeviceId}/connected
 *          子设备下线消息: /{productId}/{deviceId}/child/{childDeviceId}/disconnect
 *          读取子设备属性回复: /{productId}/{deviceId}/child/{childDeviceId}/properties/read/reply
 *          修改子设备属性回复: /{productId}/{deviceId}/child/{childDeviceId}/properties/write/reply
 *          调用子设备功能回复: /{productId}/{deviceId}/child/{childDeviceId}/function/invoke/reply
 *          上报子设备事件: /{productId}/{deviceId}/child/{childDeviceId}/event/{eventId}
 *          上报子设备派生物模型: /{productId}/{deviceId}/child/{childDeviceId}/metadata/derived
 *
 * </pre>
 * 基于jet links 的消息编解码器
 * @author zhangqiang
 * @since 0.1
 */
public enum TopicMessageCodec {

    //上报属性数据
    reportProperty("/*/properties/report", ReportPropertyMessage.class),
    //事件上报
    event("/*/event/*", EventMessage.class) {
        @Override
        Publisher<DeviceMessage> doDecode(ObjectMapper mapper, String[] topic, byte[] payload) {
            String event = topic[topic.length - 1];

            return Mono.from(super.doDecode(mapper, topic, payload))
                    .cast(EventMessage.class)
                    .doOnNext(e -> e.setEvent(event))
                    .cast(DeviceMessage.class);
        }

        @Override
        void refactorTopic(String[] topics, DeviceMessage message) {
            super.refactorTopic(topics, message);
            EventMessage event = ((EventMessage) message);
            topics[topics.length - 1] = String.valueOf(event.getEvent());
        }
    },
    //读取属性
    readProperty("/*/properties/read", ReadPropertyMessage.class),
    //读取属性回复
    readPropertyReply("/*/properties/read/reply", ReadPropertyMessageReply.class),
    //修改属性
    writeProperty("/*/properties/write", WritePropertyMessage.class),
    //修改属性回复
    writePropertyReply("/*/properties/write/reply", WritePropertyMessageReply.class),
    //调用功能
    functionInvoke("/*/function/invoke", FunctionInvokeMessage.class),
    //调用功能回复
    functionInvokeReply("/*/function/invoke/reply", FunctionInvokeMessageReply.class),
    //子设备消息
    child("/*/child/*/**", ChildDeviceMessage.class) {
        @Override
        public Publisher<DeviceMessage> doDecode(ObjectMapper mapper, String[] topic, byte[] payload) {
            String[] _topic = Arrays.copyOfRange(topic, 2, topic.length);
            _topic[0] = "";// topic以/开头所有第一位是空白
            return TopicMessageCodec
                    .decode(mapper, _topic, payload)
                    .map(childMsg -> {
                        ChildDeviceMessage msg = new ChildDeviceMessage();
                        msg.setDeviceId(topic[1]);
                        msg.setChildDeviceMessage(childMsg);
                        msg.setTimestamp(childMsg.getTimestamp());
                        msg.setMessageId(childMsg.getMessageId());
                        return msg;
                    });
        }

        @Override
        protected TopicPayload doEncode(ObjectMapper mapper, String[] topics, DeviceMessage message) {
            ChildDeviceMessage deviceMessage = ((ChildDeviceMessage) message);

            DeviceMessage childMessage = ((DeviceMessage) deviceMessage.getChildDeviceMessage());

            TopicPayload payload = TopicMessageCodec.encode(mapper, childMessage);
            String[] childTopic = payload.getTopic().split("/");
            String[] topic = new String[topics.length + childTopic.length - 3];
            //合并topic
            System.arraycopy(topics, 0, topic, 0, topics.length - 1);
            System.arraycopy(childTopic, 1, topic, topics.length - 2, childTopic.length - 1);

            refactorTopic(topic, message);
            payload.setTopic(String.join("/", topic));
            return payload;

        }
    }, //子设备消息回复
    childReply("/*/child-reply/*/**", ChildDeviceMessageReply.class) {
        @Override
        public Publisher<DeviceMessage> doDecode(ObjectMapper mapper, String[] topic, byte[] payload) {
            String[] _topic = Arrays.copyOfRange(topic, 2, topic.length);
            _topic[0] = "";// topic以/开头所有第一位是空白
            return TopicMessageCodec
                    .decode(mapper, _topic, payload)
                    .map(childMsg -> {
                        ChildDeviceMessageReply msg = new ChildDeviceMessageReply();
                        msg.setDeviceId(topic[1]);
                        msg.setChildDeviceMessage(childMsg);
                        msg.setTimestamp(childMsg.getTimestamp());
                        msg.setMessageId(childMsg.getMessageId());
                        return msg;
                    });
        }

        @Override
        protected TopicPayload doEncode(ObjectMapper mapper, String[] topics, DeviceMessage message) {
            ChildDeviceMessageReply deviceMessage = ((ChildDeviceMessageReply) message);

            DeviceMessage childMessage = ((DeviceMessage) deviceMessage.getChildDeviceMessage());

            TopicPayload payload = TopicMessageCodec.encode(mapper, childMessage);
            String[] childTopic = payload.getTopic().split("/");
            String[] topic = new String[topics.length + childTopic.length - 3];
            //合并topic
            System.arraycopy(topics, 0, topic, 0, topics.length - 1);
            System.arraycopy(childTopic, 1, topic, topics.length - 2, childTopic.length - 1);

            refactorTopic(topic, message);
            payload.setTopic(String.join("/", topic));
            return payload;

        }
    },
    //更新标签
    updateTag("/*/tags", UpdateTagMessage.class),
    //注册
    register("/*/register", DeviceRegisterMessage.class),
    //注销
    unregister("/*/unregister", DeviceUnRegisterMessage.class),
    //更新固件消息
    upgradeFirmware("/*/firmware/upgrade", UpgradeFirmwareMessage.class),
    //更新固件升级进度消息
    upgradeProcessFirmware("/*/firmware/upgrade/progress", UpgradeFirmwareProgressMessage.class),
    //拉取固件
    requestFirmware("/*/firmware/pull", RequestFirmwareMessage.class),
    //拉取固件更新回复
    requestFirmwareReply("/*/firmware/pull/reply", RequestFirmwareMessageReply.class),
    //上报固件版本
    reportFirmware("/*/firmware/report", ReportFirmwareMessage.class),
    //读取固件回复
    readFirmware("/*/firmware/read", ReadFirmwareMessage.class),
    //读取固件回复
    readFirmwareReply("/*/firmware/read/reply", ReadFirmwareMessageReply.class),
    //派生物模型上报
    derivedMetadata("/*/metadata/derived", DerivedMetadataMessage.class),
    //透传设备消息
    direct("/*/direct", DirectDeviceMessage.class) {
        @Override
        public Publisher<DeviceMessage> doDecode(ObjectMapper mapper, String[] topic, byte[] payload) {
            DirectDeviceMessage message = new DirectDeviceMessage();
            message.setDeviceId(topic[1]);
            message.setPayload(payload);
            return Mono.just(message);
        }
    },
    //断开连接消息
    disconnect("/*/disconnect", DisconnectDeviceMessage.class),
    //断开连接回复
    disconnectReply("/*/disconnect/reply", DisconnectDeviceMessageReply.class),
    //上线
    connect("/*/online", DeviceOnlineMessage.class),
    //离线
    offline("/*/offline", DeviceOfflineMessage.class),
    //日志
    log("/*/log", DeviceLogMessage.class),
    //状态检查
    stateCheck("/*/state-check", DeviceStateCheckMessage.class),
    stateCheckReply("/*/state-check/reply", DeviceStateCheckMessageReply.class),
    ;

    /**
     * 构造函数
     * @param topic
     * @param type
     */
    TopicMessageCodec(String topic, Class<? extends DeviceMessage> type) {
        this.pattern = topic.split("/");
        this.type = type;
    }

    // 主题
    private final String[] pattern;
    // DeviceMessage 类型
    private final Class<? extends DeviceMessage> type;

    /**
     * 标准解码
     * @param mapper
     * @param topics
     * @param payload
     * @return
     */
    public static Flux<DeviceMessage> decode(ObjectMapper mapper, String[] topics, byte[] payload) {
        return Mono
                .justOrEmpty(fromTopic(topics))
                .flatMapMany(topicMessageCodec -> topicMessageCodec.doDecode(mapper, topics, payload));
    }

    /**
     * 标准解码
     * @param mapper
     * @param topic
     * @param payload
     * @return
     */
    public static Flux<DeviceMessage> decode(ObjectMapper mapper, String topic, byte[] payload) {
        return decode(mapper, topic.split("/"), payload);
    }

    /**
     * 标准编码
     * @param mapper
     * @param message
     * @return
     */
    public static TopicPayload encode(ObjectMapper mapper, DeviceMessage message) {

        return fromMessage(message)
                .orElseThrow(() -> new UnsupportedOperationException("unsupported message:" + message.getMessageType()))
                .doEncode(mapper, message);
    }

    /**
     * 根据主题topic确认
     * 是否为合法码库类型
     * @param topic
     * @return
     */
    static Optional<TopicMessageCodec> fromTopic(String[] topic) {
        for (TopicMessageCodec value : values()) {
            if (TopicUtils.match(value.pattern, topic)) {
                return Optional.of(value);
            }
        }
        return Optional.empty();
    }

    /**
     * 根据消息类型型确认
     * 是否为合法码库类型
     * @param message
     * @return
     */
    static Optional<TopicMessageCodec> fromMessage(DeviceMessage message) {
        for (TopicMessageCodec value : values()) {
            if (value.type == message.getClass()) {
                return Optional.of(value);
            }
        }
        return Optional.empty();
    }

    /**
     * 标准解码
     * @param mapper
     * @param topic
     * @param payload
     * @return
     */
    Publisher<DeviceMessage> doDecode(ObjectMapper mapper, String[] topic, byte[] payload) {
        return Mono
                .fromCallable(() -> {
                    DeviceMessage message = mapper.readValue(payload, type);
                    FastBeanCopier.copy(Collections.singletonMap("deviceId", topic[1]), message);

                    return message;
                });
    }

    /**
     * 标准编码
     *
     * 默认数据结构 无需转换 平台什么数据格式就按什么数据格式编码
     * 如需自定义编码器则在重写doEncode方法即可
     * @param mapper
     * @param topics
     * @param message
     * @return
     */
    @SneakyThrows
    TopicPayload doEncode(ObjectMapper mapper, String[] topics, DeviceMessage message) {
        refactorTopic(topics, message);
        return TopicPayload.of(String.join("/", topics), mapper.writeValueAsBytes(message));
    }

    /**
     * 标准编码
     * @param mapper
     * @param message
     * @return
     */
    @SneakyThrows
    TopicPayload doEncode(ObjectMapper mapper, DeviceMessage message) {
        String[] topics = Arrays.copyOf(pattern, pattern.length);
        return doEncode(mapper, topics, message);
    }

    void refactorTopic(String[] topics, DeviceMessage message) {
        topics[1] = message.getDeviceId();
    }

    /**
     * 移除topic中的产品信息,topic第一个层为产品ID，在解码时,不需要此信息,所以需要移除之.
     *
     * @param topic topic
     * @return 移除后的topic
     */
    public static String[] removeProductPath(String topic) {
        if (!topic.startsWith("/")) {
            topic = "/" + topic;
        }
        String[] topicArr = topic.split("/");
        String[] topics = Arrays.copyOfRange(topicArr, 1, topicArr.length);
        topics[0] = "";
        return topics;
    }

}
